import torch
import torch.nn as nn
import numpy as np
from torch.utils.data import DataLoader, Subset
from abc import ABC
from typing import List, Optional, Tuple, Union, Dict
from abfml.core.model.math_fun import derive_mechanics
from abfml.param.param import Param


class FieldModel(nn.Module, ABC):
    def __init__(self,
                 type_map: List[int],
                 cutoff: float,
                 neighbor: Union[Dict[int, int],int]):
        """
        Base class for field-based machine learning potential models.

        Parameters
        ----------
        type_map : List[int]
            A list mapping atomic species to integer indices used in the model.
            For example, [0, 1, 2] might represent three different element types.

        cutoff : float
            Cutoff radius for neighbor searching.
            Only atoms within this distance will be considered as neighbors.

        neighbor : Union[Dict[int, int],int]
            Maximum number of neighbors for each atom type (or a global value).
            Used to define tensor sizes or guide model input structure.

        Example
        -------
        model = FieldModel(type_map=[0, 1], cutoff=6.0, neighbor=[1:100])
        """
        super(FieldModel, self).__init__()
        self.type_map = type_map
        self.neighbor = neighbor
        self.cutoff = cutoff

    def forward(self,
                element_map: torch.Tensor,
                central_atoms: torch.Tensor,
                neighbor_indices: torch.Tensor,
                neighbor_types: torch.Tensor,
                neighbor_vectors: torch.Tensor,
                n_ghost: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Main forward pass computing physical quantities
        Args:
            element_map: Atomic numbers mapped to indices [batch, n_atoms]
            central_atoms: Central atom indices [batch, n_atoms]
            neighbor_indices: Neighbor atom indices [batch, n_atoms, max_nbr]
            neighbor_types: Neighbor atom types [batch, n_atoms, max_nbr]
            neighbor_vectors: Relative position vectors [batch, n_atoms, max_nbr, 4]
                              (distance, dx, dy, dz)
            n_ghost: Number of ghost atoms in the system

        Returns:
            Tuple containing:
            - Etot [batch, 1]
            - Ei [batch, n_atoms, 1]
            - force [batch, n_atoms, 3]
            - virial tensor [batch, 9]
            - atomic_virial [batch, n_atoms, 9]
        """
        # Extract dimensions and tensor properties
        batch, n_atoms, max_neighbor, _ = neighbor_vectors.shape
        device, dtype = neighbor_vectors.device, neighbor_vectors.dtype

        # Extract displacement vectors (dx, dy, dz) and enable autograd
        relative_positions = neighbor_vectors[..., 1:]
        relative_positions.requires_grad_(True)

        # Recalculate distance norm for generalization and recombine into input vector
        distance = torch.norm(relative_positions, dim=-1, keepdim=True)  # Shape: [batch, n_atoms, max_nbr, 1]
        neighbor_vector = torch.cat([distance, relative_positions], dim=-1)  # Shape: [batch, n_atoms, max_nbr, 4]

        # Call the user-defined field function to compute energy and optionally other properties
        physics_info = self.field(
            element_map=element_map,
            central_atoms=central_atoms,
            neighbor_indices=neighbor_indices,
            neighbor_types=neighbor_types,
            neighbor_vectors=neighbor_vector,
            n_ghost=n_ghost
        )

        # Required outputs
        Etot = physics_info['Etot']  # Total energy per configuration, shape: [batch, 1]
        Ei = physics_info['Ei']  # Per-atom energy, shape: [batch, n_atoms, 1]

        # Validate required outputs
        if Etot is None or not isinstance(Etot, torch.Tensor) or Etot.shape != (batch, 1):
            raise ValueError("physics_info['Etot'] must be a torch.Tensor of shape (batch, 1)")
        if Ei is None or not isinstance(Ei, torch.Tensor) or Ei.shape != (batch, n_atoms, 1):
            raise ValueError("physics_info['Ei'] must be a torch.Tensor of shape (batch, n_atoms, 1)")

        # If force is not provided, compute it via autograd
        if physics_info.get('Force') is None:
            # Create a gradient mask for backpropagation
            energy_mask: List[Optional[torch.Tensor]] = [torch.ones_like(Ei)]

            # Compute gradient of Ei with respect to atomic positions
            grad_Ei = torch.autograd.grad(
                outputs=[Ei],
                inputs=[relative_positions],
                grad_outputs=energy_mask,
                retain_graph=True,
                create_graph=True
            )[0]
            assert grad_Ei is not None

            # Compute force, virial tensor, and atomic-level virial from gradients
            force, virial, atomic_virial = derive_mechanics(
                grad_Ei=grad_Ei,
                neighbor_vectors=neighbor_vector,
                neighbor_indices=neighbor_indices,
                n_ghost=n_ghost
            )
        else:
            # If force is provided by field, validate its shape
            force = physics_info['Force']
            if not isinstance(force, torch.Tensor) or force.shape != (batch, n_atoms, 3):
                raise ValueError("physics_info['Force'] must be a torch.Tensor of shape (batch, n_atoms, 3)")

            # Optional: check for virial tensor
            virial = physics_info.get('virial')
            if virial is not None:
                if not isinstance(virial, torch.Tensor) or virial.shape != (batch, 3, 3):
                    raise ValueError("physics_info['virial'] must be a torch.Tensor of shape (batch, 3, 3)")
            else:
                virial = torch.zeros(batch, 3, 3, dtype=dtype, device=device)

            # Optional: check for atomic-level virial
            atomic_virial = physics_info.get('atomic_virial')
            if atomic_virial is not None:
                if not isinstance(atomic_virial, torch.Tensor) or atomic_virial.shape != (batch, n_atoms, 9):
                    raise ValueError(
                        "physics_info['atomic_virial'] must be a torch.Tensor of shape (batch, n_atoms, 9)")
            else:
                atomic_virial = torch.zeros(batch, n_atoms, 9, dtype=dtype, device=device)

        # Return all computed quantities
        return Etot, Ei, force, virial, atomic_virial

    def field(self,
              element_map: torch.Tensor,
              central_atoms: torch.Tensor,
              neighbor_indices: torch.Tensor,
              neighbor_types: torch.Tensor,
              neighbor_vectors: torch.Tensor,
              n_ghost: int) -> Dict[str, torch.Tensor]:
        """
        Abstract method to be implemented by subclasses. This method defines the core
        field calculation of the model, which computes atom-wise energies and optionally
        forces and virial tensors based on atomic environments.

        Parameters
        ----------
        element_map : torch.Tensor
            Tensor of shape [batch, n_atoms], indicating the atomic species (element index)
            for each atom in the batch.

        central_atoms : torch.Tensor
            Tensor of shape [batch, n_atoms], representing the type of central atoms.

        neighbor_indices : torch.Tensor
            Tensor of shape [batch, n_atoms, max_nbr], providing the indices of neighboring
            atoms for each central atom.

        neighbor_types : torch.Tensor
            Tensor of shape [batch, n_atoms, max_nbr], indicating the atomic species
            of neighboring atoms.

        neighbor_vectors : torch.Tensor
            Tensor of shape [batch, n_atoms, max_nbr, 4], where the last dimension represents
            (distance, dx, dy, dz), i.e., the norm and vector components of the relative positions.

        n_ghost : int
            Number of ghost atoms (used to pad or augment neighbor lists, typically for parallelism).

        Returns
        -------
        Dict[str, torch.Tensor]
            A dictionary containing the computed physical quantities:
            - 'Etot': Total energy, shape [batch, 1].
            - 'Ei': Atomic energy per atom, shape [batch, n_atoms, 1].
            - Optional: 'Force': Force on each atom, shape [batch, n_atoms, 3].
            - Optional: 'virial': Virial tensor, shape [batch, 9].
            - Optional: 'atomic_virial': virial Force on each atom, shape [batch, n_atoms, 9].

        Note
        ----
        This method must be implemented in subclasses of `FieldModel`.
        """
        return {'Etot': torch.tensor([0.0])}


class NormalModel:
    def __init__(self,
                 normal_data,
                 param_class: Param,
                 normal_rate: Union[float, str] = 'auto',
                 is_get_energy_shift: bool = False):
        self.param_class = param_class
        self.normal_loader = NormalModel.normal_data_loader(need_data=normal_data, normal_rate=normal_rate)
        self.is_get_energy_shift = is_get_energy_shift
        self.normal_data = tuple([])

    def initialize(self):
        normal_data = self.normal(normal_loader=self.normal_loader, param_class=self.param_class)

        if isinstance(normal_data, tuple):
            self.normal_data = normal_data
        else:
            self.normal_data = tuple([normal_data])

        if self.is_get_energy_shift:
            energy_shift = NormalModel.get_energy_shift(need_data=self.normal_loader, type_map=self.param_class.GlobalSet.type_map)
            self.normal_data = tuple([energy_shift]) + self.normal_data

    @staticmethod
    def normal_data_loader(need_data, normal_rate: Union[float, str]) -> DataLoader:
        total_image_num = len(need_data)
        total_indices = np.arange(total_image_num)
        if isinstance(normal_rate, float):
            if normal_rate <= 1.0:
                num = int(total_image_num * normal_rate + 1)
            else:
                raise Exception("rate")
        elif normal_rate == "auto":
            if total_image_num * 0.1 < 100:
                num = total_image_num
            else:
                num = int(total_image_num * 0.1 + 1)
        else:
            raise Exception("rate")
        np.random.shuffle(total_indices)
        normal_indices = total_indices[:num]
        normal_data = Subset(need_data, normal_indices)
        num_threads = torch.get_num_threads()
        num_worker = int(num_threads / 2)
        normal_data_loader = DataLoader(normal_data, batch_size=1, shuffle=True, num_workers=num_worker)

        return normal_data_loader

    @staticmethod
    def get_energy_shift(need_data, type_map: List[int]) -> List[float]:
        ntype = len(type_map)
        type_num = torch.zeros(ntype)
        energy_shift = [0.0] * ntype
        for i, image_batch in enumerate(need_data):
            central_atoms = image_batch["central_atoms"]
            element_types = image_batch["element_types"][0].to(torch.int64).tolist()
            for i_type, element in enumerate(element_types):
                mask = (central_atoms == element)
                indices = type_map.index(element)
                type_num[indices] += 1
                try:
                    energy = torch.mean(image_batch["energy"] / image_batch["n_atoms"]).item()
                    energy_shift[indices] = energy_shift[indices] + energy
                except KeyError:
                    try:
                        Ei = torch.mean(image_batch["atomic_energy"][mask]).item()
                        energy_shift[indices] = energy_shift[indices] + Ei
                    except KeyError:
                        energy_shift[indices] = energy_shift[indices] + np.random.uniform(-10.0, 0.0)

        type_num[type_num == 0] = 1
        for i, i_energy in enumerate(energy_shift):
            energy_shift[i] = (i_energy / type_num[i]).item()

        return energy_shift

    def normal(self, normal_loader, param_class):
        return None




import torch
import numpy as np
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from ase import Atoms, io
from ase.neighborlist import NeighborList
from typing import List, Optional, Union, Dict
from torch_geometric.data import Data
from dataclasses import dataclass
from abfml.data.dataset import ABFMLDataset


@dataclass
class DataClass:
    filename: str
    atoms_list: List[Atoms]
    include_element: set = None
    include_atoms_numbers: set = None
    num_frames: int = 0
    def __init__(self, filename: str, file_format: Optional[str]):
        self.filename: str = filename
        if file_format == "pwmat-config":
            atoms_list = []
        elif file_format == "pwmat-movement":
            atoms_list = []
        else:
            atoms_list = io.read(filename, format=file_format, index=':')
        self.atoms_list: List[Atoms] = atoms_list
        self.include_element: set = set()
        include_atoms_number = []
        for atom_information in atoms_list:
            self.include_element = self.include_element.union(set(atom_information.get_chemical_symbols()))
            include_atoms_number.append(len(atom_information))
        self.include_atoms_number: set = set(include_atoms_number)
        self.n_frames: int = len(self.atoms_list)


class ReadData:
    def __init__(
            self,
            filename: Union[str, List[str]],
            cutoff: float,
            neighbor: Union[Dict[int, int], int],
            file_format: Optional[str]):
        self.cutoff = cutoff
        self.neighbor = neighbor
        self.filename = filename
        self.atoms_list: list[Atoms] = []
        if isinstance(filename, str):
            filename = [filename]
        self.data_information: List[dict] = []
        for file in filename:
            data_information = DataClass(file, file_format)
            self.data_information.append({'file_name': data_information.filename,
                                          'n_frames': data_information.n_frames,
                                          'include_element': data_information.include_element,
                                          'include_atoms_number': data_information.include_atoms_number})
            self.atoms_list = self.atoms_list + data_information.atoms_list
        self.n_frames: int = len(self.atoms_list)
        self.unique_image = {}
        for i, image in enumerate(self.atoms_list):
            image_tuple = tuple(image.get_atomic_numbers())
            if image_tuple in self.unique_image:
                self.unique_image[image_tuple].append(i)
            else:
                self.unique_image[image_tuple] = [i]

    def create_dataset(self):
        dataset = []
        for same_image_index in self.unique_image.values():
            n_frames: int = len(same_image_index)
            n_atoms: int = len(self.atoms_list[same_image_index[0]])
            n_elements: int = len(set(self.atoms_list[same_image_index[0]].get_atomic_numbers()))
            if isinstance(self.neighbor, int):
                max_neighbor: int = self.neighbor
            elif isinstance(self.neighbor, dict):
                max_neighbor = sum(self.neighbor[atomic_numbers]
                                       for atomic_numbers in set(self.atoms_list[same_image_index[0]].get_atomic_numbers()))
            else:
                raise ValueError(f'Expected a member of Dict[int, int] and int but instead found type {type(self.neighbor)}')

            central_atoms = torch.empty(n_frames, n_atoms, dtype=torch.int)
            neighbor_indices = torch.empty(n_frames, n_atoms, max_neighbor, dtype=torch.int)
            neighbor_types = torch.empty(n_frames, n_atoms, max_neighbor, dtype=torch.int)
            neighbor_vectors = torch.zeros(n_frames, n_atoms, max_neighbor, 4, dtype=torch.float64)
            element_types = torch.zeros(n_frames, n_elements, dtype=torch.int)
            energy = torch.zeros(n_frames, 1, dtype=torch.float64)
            atomic_energies = torch.zeros(n_frames, n_atoms, 1, dtype=torch.float64)
            forces = torch.zeros(n_frames, n_atoms, 3, dtype=torch.float64)
            virials = torch.zeros(n_frames, 9, dtype=torch.float64)
            for i, image_index in enumerate(same_image_index):
                atoms_ase = self.atoms_list[image_index]
                neighbor_information = ReadData.calculate_neighbor(atoms=atoms_ase,
                                                                   cutoff=self.cutoff,
                                                                   neighbor=self.neighbor)
                element_types = torch.tensor(neighbor_information[0])
                central_atoms[i] = torch.tensor(neighbor_information[1])
                neighbor_indices[i] = torch.tensor(neighbor_information[2])
                neighbor_types[i] = torch.tensor(neighbor_information[3])
                neighbor_vectors[i] = torch.tensor(neighbor_information[4])
                try:
                    energy[i] = torch.tensor(atoms_ase.get_potential_energy())
                except (RuntimeError, AttributeError):
                    energy = None
                try:
                    atomic_energies[i] = torch.tensor(atoms_ase.atomic_energies)
                except (RuntimeError, AttributeError):
                    atomic_energies = None
                try:
                    forces[i] = torch.tensor(atoms_ase.get_forces(apply_constraint=False))
                except (RuntimeError, AttributeError):
                    forces = None
                try:
                    # ASE calculated in units of eV/A^3, virial: eV
                    virials[i] = torch.tensor(-1.0 * atoms_ase.get_stress(voigt=False).reshape(9) * atoms_ase.get_volume())
                except (RuntimeError, AttributeError):
                    virials = None
            dataset.append(ABFMLDataset(num_frames=n_frames,
                                        num_atoms=n_atoms,
                                        element_types=element_types,
                                        central_atoms=central_atoms,
                                        neighbor_indices=neighbor_indices,
                                        neighbor_types=neighbor_types,
                                        neighbor_vectors=neighbor_vectors,
                                        energy=energy,
                                        atomic_energies=atomic_energies,
                                        forces=forces,
                                        virials=virials))
        return dataset

    @staticmethod
    def calculate_neighbor(atoms: Atoms,
                           cutoff: float,
                           neighbor: Union[Dict[int, int], int]):
        nl = NeighborList([cutoff / 2] * len(atoms), skin=0, self_interaction=False, bothways=True, sorted=False)
        nl.update(atoms)
        atoms_num = len(atoms)
        central_atoms = atoms.numbers
        element_types = np.unique(central_atoms)
        if isinstance(neighbor, int):
            max_neighbor = neighbor
            neighbor_vectors = np.zeros(shape=(1, atoms_num, max_neighbor, 4), dtype=np.float64)
            neighbor_indices = np.full(shape=(1, atoms_num, max_neighbor), fill_value=-1, dtype=np.int32)
            neighbor_types = np.full(shape=(1, atoms_num, max_neighbor), fill_value=-1, dtype=np.int32)
            for i in range(atoms_num):
                indices_neighbor, offsets = nl.get_neighbors(i)
                neighbor_positions = atoms.positions[indices_neighbor] + np.dot(offsets, atoms.get_cell())
                delta = neighbor_positions - atoms.positions[i]
                rij = np.linalg.norm(delta, axis=1)
                indices = np.arange(indices_neighbor.shape[0])
                if indices_neighbor.shape[0] <= max_neighbor:
                    sorted_indices = indices
                else:
                    order = np.argsort(rij, axis=-1)
                    sorted_indices = indices[order[:max_neighbor]]
                valid_count = len(sorted_indices)

                neighbor_vectors[:, i, :valid_count, 1:] = delta[sorted_indices, :]
                neighbor_vectors[:, i, :valid_count, 0] = rij[sorted_indices]
                neighbor_indices[:, i, :valid_count] = indices_neighbor[sorted_indices]
            mask = (neighbor_indices != -1)
            neighbor_types[mask] = central_atoms[neighbor_indices[mask]]

        elif isinstance(neighbor, Dict):
            max_neighbor = sum(neighbor[element] for element in element_types)
            slot_map = {}
            current_slot = 0
            for atomic_number in neighbor.keys():
                slot_size = neighbor[atomic_number]
                slot_map[atomic_number] = (current_slot, current_slot + slot_size)
                current_slot += slot_size
            central_atoms = atoms.numbers
            neighbor_vectors = np.zeros(shape=(1, atoms_num, max_neighbor, 4), dtype=np.float64)
            neighbor_indices = np.full(shape=(1, atoms_num, max_neighbor), fill_value=-1, dtype=np.int32)
            neighbor_types = np.full(shape=(1, atoms_num, max_neighbor), fill_value=-1, dtype=np.int32)
            for i in range(atoms_num):
                indices_neighbor, offsets = nl.get_neighbors(i)
                neighbor_positions = atoms.positions[indices_neighbor] + np.dot(offsets, atoms.get_cell())
                delta = neighbor_positions - atoms.positions[i]
                rij = np.linalg.norm(delta, axis=1)

                for atomic_numbers in np.unique(element_types):
                    mask = (central_atoms[indices_neighbor])==atomic_numbers
                    if not np.any(mask):
                        continue
                    start_slot, end_slot = slot_map[atomic_numbers]
                    max_slots = end_slot - start_slot
                    valid_indices = np.where(mask)[0]
                    valid_rij = rij[valid_indices]
                    if valid_indices.shape[0] <= max_slots:
                        sorted_indices = valid_indices
                    else:
                        order = np.argsort(valid_rij, axis=-1)
                        sorted_indices = valid_indices[order[:max_slots]]
                    valid_count = len(sorted_indices)
                    slot_slice = slice(start_slot, start_slot + valid_count)
                    neighbor_vectors[:, i, slot_slice, 1:] = delta[sorted_indices, :]
                    neighbor_vectors[:, i, slot_slice, 0] = rij[sorted_indices]
                    neighbor_indices[:, i, slot_slice] = indices_neighbor[sorted_indices]
            mask = (neighbor_indices != -1)
            neighbor_types[mask] = central_atoms[neighbor_indices[mask]]
        else:
            raise Exception('neighbor[0] and neighbor[1] should have the same length '
                            'or neighbor[1] have only one element. Maybe you should read the manual!')
        central_atoms = central_atoms[np.newaxis, ...]
        return element_types, central_atoms, neighbor_indices, neighbor_types, neighbor_vectors


def to_graph(central_atoms: torch.Tensor,
             neighbor_indices: torch.Tensor,
             neighbor_types: torch.Tensor,
             neighbor_vectors: torch.Tensor) -> List[Data]:
    """
    Convert atomic neighbor information into a list of PyG graph objects (one per batch entry).

    Args:
        central_atoms (Tensor): Shape (B, N), atomic types of central atoms.
        neighbor_indices (Tensor): Shape (B, N, M), indices of neighbors (-1 for invalid).
        neighbor_types (Tensor): Shape (B, N, M), types of neighboring atoms.
        neighbor_vectors (Tensor): Shape (B, N, M, 4), relative vectors to neighbors (r, dx, dy, dz).

    Notes: For the safety of subsequent operations, this out returns a list instead of a Batch of torch_geometric.data.
    Returns:
        List[Data]: A list of torch_geometric.data.Data graphs, one for each batch item.
        Each Data object contains:
            - x (Tensor): Node features, shape [num_nodes, num_node_features]. Here, it's the atomic type.
            - edge_index (LongTensor): Graph connectivity in COO format, shape [2, num_edges].
            - relative_pos (Tensor): Edge features (e.g., distances and relative positions), shape [num_edges, 4].
    """
    batch, n_atoms, max_neighbor = neighbor_types.shape
    device, dtype = neighbor_vectors.device, neighbor_vectors.dtype
    graphs = []

    for b in range(batch):
        # Create node features from central atom types (shape: [N, 1])
        x = central_atoms[b].unsqueeze(-1).to(torch.long)

        # Identify valid neighbor connections (exclude -1 entries)
        valid_mask = neighbor_indices[b] != -1

        # Get valid (source_atom_index, neighbor_index_in_M) pairs
        src_idx, nbr_idx = valid_mask.nonzero(as_tuple=True)

        # Source atoms are the indices from 0 to N-1 (repeated)
        edge_src = src_idx

        # Destination atoms are the valid neighbor indices
        edge_dst = neighbor_indices[b][src_idx, nbr_idx]

        # Stack to form edge_index of shape [2, num_edges]
        edge_index = torch.stack([edge_src, edge_dst], dim=0).to(device)

        # Extract relative vectors (r, dx, dy, dz)
        relative_pos = neighbor_vectors[b][src_idx, nbr_idx].to(device=device, dtype=dtype)
        # Construct a PyG Data object for this graph
        graph = Data(x=x, edge_index=edge_index, relative_pos=relative_pos)
        graphs.append(graph)
    return graphs





import torch
from torch.utils.data import Dataset


class ABFMLDataset(Dataset):
    def __init__(
        self,
        num_frames: int,
        num_atoms: int,
        element_types: torch.Tensor,
        central_atoms: torch.Tensor,
        neighbor_indices: torch.Tensor,
        neighbor_types: torch.Tensor,
        neighbor_vectors: torch.Tensor,
        energy: torch.Tensor = None,
        atomic_energies: torch.Tensor = None,
        forces: torch.Tensor = None,
        virials: torch.Tensor = None,
    ):
        super().__init__()
        self.num_frames = num_frames
        self.num_atoms = num_atoms

        # Validate physical quantities
        physics_info = {
            "energy": (energy, (num_frames, 1)),
            "atomic_energies": (atomic_energies, (num_frames, num_atoms, 1)),
            "forces": (forces, (num_frames, num_atoms, 3)),
            "virials": (virials, (num_frames, 9)),
        }

        for name, (tensor, expected_shape) in physics_info.items():
            if tensor is not None:
                if not isinstance(tensor, torch.Tensor):
                    raise TypeError(f"{name} must be a torch.Tensor, but got {type(tensor)}.")
                if tensor.shape != expected_shape:
                    raise ValueError(f"{name} must have shape {expected_shape}, but got {tensor.shape}.")

        self.energy = energy
        self.atomic_energies = atomic_energies
        self.forces = forces
        self.virials = virials

        # Validate neighbor-related tensors
        num_neighbors = neighbor_indices.shape[-1]
        neighbor_info = {
            "central_atoms": (central_atoms, (num_frames, num_atoms)),
            "neighbor_indices": (neighbor_indices, (num_frames, num_atoms, num_neighbors)),
            "neighbor_types": (neighbor_types, (num_frames, num_atoms, num_neighbors)),
            "neighbor_vectors": (neighbor_vectors, (num_frames, num_atoms, num_neighbors, 4)),
        }

        for name, (tensor, expected_shape) in neighbor_info.items():
            if not isinstance(tensor, torch.Tensor):
                raise TypeError(f"{name} must be a torch.Tensor, but got {type(tensor)}.")
            if tensor.shape != expected_shape:
                raise ValueError(f"{name} must have shape {expected_shape}, but got {tensor.shape}.")

        self.element_types = element_types
        self.central_atoms = central_atoms
        self.neighbor_indices = neighbor_indices
        self.neighbor_types = neighbor_types
        self.neighbor_vectors = neighbor_vectors

    def __len__(self):
        return self.num_frames

    def __getitem__(self, index):
        sample = {
            "num_atoms": self.num_atoms,
            "element_types": self.element_types,
            "central_atoms": self.central_atoms[index],
            "neighbor_indices": self.neighbor_indices[index],
            "neighbor_types": self.neighbor_types[index],
            "neighbor_vectors": self.neighbor_vectors[index],
        }

        if self.energy is not None:
            sample["energy"] = self.energy[index]
        if self.atomic_energies is not None:
            sample["atomic_energies"] = self.atomic_energies[index]
        if self.forces is not None:
            sample["forces"] = self.forces[index]
        if self.virials is not None:
            sample["virials"] = self.virials[index]

        return sample



# SPDX-License-Identifier: LGPL-3.0-or-later
"""Register entry points for ABFML with LAMMPS."""

import os
import platform
from importlib import import_module
from pathlib import Path
from typing import Optional

if platform.system() == "Linux":
    lib_env = "LD_LIBRARY_PATH"
    preload_env = "LD_PRELOAD"
elif platform.system() == "Darwin":
    lib_env = "DYLD_FALLBACK_LIBRARY_PATH"
    preload_env = "DYLD_INSERT_LIBRARIES"
else:
    raise RuntimeError("Unsupported platform")


def get_env(paths: list[Optional[str]]) -> str:
    """Get the environment variable from given paths."""
    return ":".join(p for p in paths if p is not None)


def get_library_path(module: str, filename: str) -> list[str]:
    """Get library path from a module.

    Parameters
    ----------
    module : str
        The module name.
    filename : str
        The library filename pattern.

    Returns
    -------
    list[str]
        The library path.
    """
    try:
        m = import_module(module)
    except ModuleNotFoundError:
        return []
    else:
        libs = sorted(Path(m.__path__[0]).glob(filename))
        return [str(lib) for lib in libs]


# Get PyTorch C++ extension directory
try:
    import torch

    torch_dir = torch.utils.cmake_prefix_path
except ImportError:
    raise RuntimeError("PyTorch is not installed!")

# Set paths for ABFML
abfml_lib_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "lib"))


# Set environment variables
os.environ[preload_env] = get_env([
    os.environ.get(preload_env)
])

os.environ[lib_env] = get_env([
    os.environ.get(lib_env),
    torch_dir,
    abfml_lib_dir,
])


def get_abfml_lib_dir() -> str:
    """Get the directory of the ABFML library."""
    return abfml_lib_dir



import math
from abfml.param.param import LrSet
from abfml.logger.loggers import Logger


def adjust_lr(lr_param: LrSet, iters_step: int) -> float:
    """Calculate the current learning rate based on schedule type."""
    if lr_param.type_name == 'exp':
        lr = lr_param.start_lr * lr_param.decay_rate ** (iters_step / lr_param.decay_step)
    elif lr_param.type_name == 'exp_decay':
        lr = lr_param.start_lr * lr_param.decay_rate ** (iters_step // lr_param.decay_step)
    else:
        raise ValueError(f"Unsupported learning rate type: {lr_param.type_name}")

    return max(lr, lr_param.limit_lr)


def calc_lr_param(lr_param: LrSet, epochs: int, train_iters: int, logger_name: str = None):
    """
    Calculate the decay_rate if not set. Otherwise estimate when limit_lr will be reached.
    """
    total_iters = epochs * train_iters
    logger = Logger(logger_name).logger if logger_name else None

    if lr_param.decay_rate is None:
        # Compute decay_rate to reach limit_lr in total_iters
        if total_iters > 0:
            lr_param.decay_rate = (lr_param.limit_lr / lr_param.start_lr) ** (lr_param.decay_step / total_iters)
            if logger:
                logger.info(f"[Auto LR] start_lr: {lr_param.start_lr:.2e}, limit_lr: {lr_param.limit_lr:.2e}, "
                            f"decay_step: {lr_param.decay_step}, computed decay_rate: {lr_param.decay_rate:.4f}")
        else:
            raise ValueError("Total training iterations (epochs * train_iters) must be greater than zero.")
    else:
        # Estimate how many iterations are needed to reach limit_lr
        try:
            needed_iters = math.log(lr_param.limit_lr / lr_param.start_lr) / math.log(lr_param.decay_rate)
            needed_iters *= lr_param.decay_step
            estimated_epochs = math.ceil(needed_iters / train_iters) if train_iters > 0 else float('inf')

            if logger:
                if needed_iters < total_iters:
                    logger.info(f"[LR Info] limit_lr will be reached in ~{estimated_epochs} epochs.")
                else:
                    logger.warning(f"[LR Warning] Need ~{estimated_epochs} epochs to reach limit_lr, "
                                   f"but training is set to {epochs} epochs.")
        except (ValueError, ZeroDivisionError) as e:
            if logger:
                logger.error(f"[LR Error] Failed to compute decay trajectory: {e}")




import argparse
from typing import List, Optional, Union
from abfml.entrypoints.train import train_mlff
from abfml.entrypoints.valid import valid_mlff
from abfml.entrypoints.check import check_mlff

def main_command(args: Optional[Union[List[str], argparse.Namespace]] = None):
    dict_args = vars(args)
    if args.command == "train":
        train_mlff(**dict_args)
    elif args.command == "valid":
        valid_mlff(**dict_args)
    elif args.command == "check":
        check_mlff(**dict_args)
    else:
        raise Exception(f"undefined command")



from collections import OrderedDict
from abfml.param.param import LossSet
from abfml.logger.loggers import AverageMeter


def calculate_weight(param: LossSet, learn_rate: float):
    attributes = ['energy',  'ei', 'force', 'virial']
    weights = []

    for attr in attributes:
        start_weight = getattr(param, f'start_{attr}_weight')
        limit_weight = getattr(param, f'limit_{attr}_weight')
        weight = limit_weight * (1 - learn_rate) + start_weight * learn_rate
        weights.append(weight)

    return tuple(weights)


def register_loss_meters(loss_set: LossSet = None, base_meters=True):
    """
       General-purpose meter registration function.

       Parameters:
           loss_set (object): A config object that contains start and limit weights for each loss term.
           base_meters (bool): Whether to include basic meters like 'loss' and 'time'.

       Returns:
           OrderedDict: A dictionary of AverageMeter instances keyed by loss type.
       """
    meters = OrderedDict()

    if base_meters:
        meters["loss"] = AverageMeter("Loss", fmt=".2e", summary_type="AVERAGE")
        meters["time"] = AverageMeter("Time", fmt="2.3f", summary_type="SUM")

    if loss_set is not None:
        # Register optional loss terms based on non-zero weights
        optional_losses = {
            "energy": ("E_tot", loss_set.start_energy_weight, loss_set.limit_energy_weight),
            "atomic_energies": ("Ei", loss_set.start_ei_weight, loss_set.limit_ei_weight),
            "forces": ("Force", loss_set.start_force_weight, loss_set.limit_force_weight),
            "virials": ("Virial", loss_set.start_virial_weight, loss_set.limit_virial_weight)
        }

        for key, (display_name, start, limit) in optional_losses.items():
            if start != 0.0 or limit != 0.0:
                meters[key] = AverageMeter(display_name, fmt=".2e", summary_type="AVERAGE")

    else:
        display_names = {
            "energy": "E_tot",
            "atomic_energies": "Ei",
            "forces": "Force",
            "virials": "Virial"
        }
        for key in display_names:
            meters[key] = AverageMeter(display_names[key], fmt=".2e", summary_type="AVERAGE")
    return meters



import torch
import torch.nn as nn
from typing import List, Tuple


class ActivationModule(nn.Module):
    def __init__(self, activation_name: str):
        super(ActivationModule, self).__init__()
        activation_functions = {
            'relu': nn.ReLU(),
            'tanh': nn.Tanh(),
            'sigmoid': nn.Sigmoid(),
            'softplus': nn.Softplus(),
            'softsign': nn.Softsign(),
            'elu': nn.ELU(),
            'leaky_relu': nn.LeakyReLU(),
            'prelu': nn.PReLU(),
            'selu': nn.SELU(),
            'gelu': nn.GELU(),
            'celu': nn.CELU(),
            'glu': nn.GLU(),
            'logsigmoid': nn.LogSigmoid(),
            'rrelu': nn.RReLU(),
            'hardshrink': nn.Hardshrink(),
            'hardtanh': nn.Hardtanh(),
            'softshrink': nn.Softshrink(),
            'tanhshrink': nn.Tanhshrink()
        }
        self.activation = activation_functions.get(activation_name, None)

    def forward(self, x):
        return self.activation(x)


@torch.jit.script
def smooth_fun(smooth_type: str, rij: torch.Tensor, r_inner: float, r_outer: float) -> torch.Tensor:
    fx = torch.zeros_like(rij, dtype=rij.dtype, device=rij.device)
    if smooth_type == 'cos':
        mask = (rij > 1e-5) & (rij < r_outer)
        fx[mask] = (0.5 * torch.cos(torch.pi * rij[mask] / r_outer) + 0.5)
    elif smooth_type == 'cos_r':
        fx[(rij > 1e-5) & (rij < r_inner)] = 1 / rij[(rij > 1e-5) & (rij < r_inner)]
        mask = (rij > r_inner) & (rij < r_outer)
        x = (rij[mask] - r_inner) / (r_outer - r_inner)
        fx[mask] = (0.5 * torch.cos(torch.pi * x) + 0.5) / rij[mask]
    elif smooth_type == 'tanh_u':
        fx[(rij > 1e-5) & (rij < r_inner)] = 1 / rij[(rij > 1e-5) & (rij < r_inner)]
        mask = (rij > r_inner) & (rij < r_outer)
        x = (rij[mask] - r_inner) / (r_outer - r_inner)
        fx[mask] = torch.tanh(1 - x) ** 3
    elif smooth_type == 'exp':
        mask = (rij > 1e-5) & (rij < r_outer)
        x = (rij[mask] - r_inner) / (r_outer - r_inner)
        fx[mask] = torch.exp(1-1/(1-torch.square(x)))
    elif smooth_type == 'poly1':
        mask = (rij > 1e-5) & (rij < r_outer)
        x = (rij[mask] - r_inner) / (r_outer - r_inner)
        fx[mask] = ((2 * x - 3) * (x ** 2) + 1)
    elif smooth_type == 'poly2':
        mask = (rij > 1e-5) & (rij < r_outer)
        x = (rij[mask] - r_inner) / (r_outer - r_inner)
        fx[mask] = ((-6 * x ** 2 + 15 * x - 10) * (x ** 3) + 1)
    elif smooth_type == 'poly3':
        mask = (rij > 1e-5) & (rij < r_outer)
        x = (rij[mask] - r_inner) / (r_outer - r_inner)
        fx[mask] = ((20 * x ** 3 - 70 * x ** 2 + 84 * x - 35) * (x ** 4) + 1)
    elif smooth_type == 'poly4':
        mask = (rij > 1e-5) & (rij < r_outer)
        x = (rij[mask] - r_inner) / (r_outer - r_inner)
        fx[mask] = ((- 70 * x ** 4 + 315 * x ** 3 - 540 * x ** 2 + 420 * x - 126) * (x ** 5) + 1)
    elif smooth_type == 'poly1_r':
        fx[(rij > 1e-5) & (rij < r_inner)] = 1 / rij[(rij > 1e-5) & (rij < r_inner)]
        mask = (rij > r_inner) & (rij < r_outer)
        x = (rij[mask] - r_inner) / (r_outer - r_inner)
        fx[mask] = ((2 * x - 3) * (x ** 2) + 1) / rij[mask]
    elif smooth_type == 'poly2_r':
        fx[(rij > 1e-5) & (rij < r_inner)] = 1 / rij[(rij > 1e-5) & (rij < r_inner)]
        mask = (rij > r_inner) & (rij < r_outer)
        x = (rij[mask] - r_inner) / (r_outer - r_inner)
        fx[mask] = ((-6 * x ** 2 + 15 * x - 10) * (x ** 3) + 1) / rij[mask]
    elif smooth_type == 'poly3_r':
        fx[(rij > 1e-5) & (rij < r_inner)] = 1 / rij[(rij > 1e-5) & (rij < r_inner)]
        mask = (rij > r_inner) & (rij < r_outer)
        x = (rij[mask] - r_inner) / (r_outer - r_inner)
        fx[mask] = ((20 * x ** 3 - 70 * x ** 2 + 84 * x - 35) * (x ** 4) + 1) / rij[mask]
    else:
        raise KeyError(f'Undefined smooth types {smooth_type}')
    return fx


@torch.jit.script
def polynomial_fun(fun_name: str, n: int, rij: torch.Tensor, r_inner: float, r_outer: float) -> torch.Tensor:
    shape = list(rij.shape[:-1]) + [n + 1]
    fx = torch.zeros(shape, dtype=rij.dtype, device=rij.device)
    if n < 2:
        raise ValueError('n must be greater than 2')
    if fun_name == 'chebyshev':
        mask = (rij > r_inner) & (rij < r_outer)
        x = 2 * (rij[mask] - r_inner) / (r_outer - r_inner) - 1
        fx[..., 0:1][mask] = 1
        fx[..., 1:2][mask] = x
        for i in range(1, n + 1):
            fx_temp_1 = (2 * fx[..., 1]).clone()
            fx_temp_2 = fx[..., i - 1].clone()
            fx_temp_3 = fx[..., i - 2].clone()
            fx[..., i] = fx_temp_1 * fx_temp_2 - fx_temp_3
    return fx


#@torch.jit.script
def derive_mechanics(
    grad_Ei: torch.Tensor,
    neighbor_vectors: torch.Tensor,
    neighbor_indices: torch.Tensor,
    n_ghost: int
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    batch, n_atoms, max_neighbors = neighbor_indices.shape
    dtype, device = neighbor_vectors.dtype, neighbor_vectors.device

    # Initialize force and per-atom virial tensors
    force = torch.zeros(batch, n_atoms + n_ghost, 3, dtype=dtype, device=device)
    force[:, :n_atoms, :] = grad_Ei.sum(dim=-2)  # Direct contribution from central atoms

    atomic_virial = torch.zeros(batch, n_atoms + n_ghost, 9, dtype=dtype, device=device)

    # Extract relative position vectors (exclude norm)
    rel_pos = neighbor_vectors[..., 1:]  # Shape: [batch, n_atoms, max_neighbors, 3]

    # Compute local virial contributions via outer product of dr and dE
    local_virials = torch.matmul(rel_pos.unsqueeze(-1), -1 * grad_Ei.unsqueeze(-2)).reshape(batch, n_atoms, max_neighbors, 9)

    # Replace invalid neighbor indices (-1) with 0
    neighbor_indices[neighbor_indices == -1] = 0

    for b in range(batch):
        # Flatten indices and contribution values
        neighbor_idx_b = neighbor_indices[b].view(-1).to(torch.int64)

        # Force scatter-add
        force_contrib_b = -1 * grad_Ei[b].view(-1, 3)
        index_f = neighbor_idx_b.unsqueeze(-1).expand(-1, 3)
        force[b] = force[b].scatter_add(0, index_f, force_contrib_b)

        # Virial scatter-add
        virial_contrib_b = local_virials[b].view(-1, 9)
        index_v = neighbor_idx_b.unsqueeze(-1).expand(-1, 9)
        atomic_virial[b] = atomic_virial[b].scatter_add(0, index_v, virial_contrib_b)

    # Compute total virial by summing over all atoms
    virial = atomic_virial.sum(dim=1)

    return force, virial, atomic_virial




import ase
import torch
from typing import List
from ase.calculators.calculator import Calculator, all_changes, PropertyNotImplementedError
from abfml.data.read_data import ReadData


class ABFML(Calculator):
    implemented_properties = ['energy', 'energies', 'forces', 'stress', 'stresses']

    def __init__(self, model: str = None, dtype: str = 'float64', model_class=None, **kwargs):
        super().__init__(**kwargs)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.dtype = torch.float32 if dtype == "float32" else torch.float64

        self.model = self._load_model(model, model_class)
        self.model.to(self.dtype).to(self.device).eval()

    def _load_model(self, model, model_class):
        if model:
            try:
                return torch.jit.load(model, map_location=self.device)
            except Exception:
                try:
                    return torch.load(model, map_location=self.device)
                except Exception as e:
                    raise RuntimeError(f"Failed to load model '{model}' with jit and torch.load: {e}")
        elif model_class:
            return model_class
        else:
            raise RuntimeError("No model or model_class provided.")

    def calculate(self, atoms: ase.Atoms = None, properties: List[str] = None, system_changes=all_changes):
        super().calculate(atoms, properties, system_changes)

        props = properties or self.implemented_properties
        model = self.model
        type_map = model.type_map
        cutoff = model.cutoff
        neighbor = model.neighbor

        # Get neighbor info
        element_type, central_atoms, neighbor_indices, neighbor_types, neighbor_vectors = \
            ReadData.calculate_neighbor(atoms=atoms, cutoff=cutoff, neighbor=neighbor)

        # Prepare inputs
        input_tensors = [
            torch.tensor(element_type),
            torch.tensor(central_atoms, device=self.device),
            torch.tensor(neighbor_indices, device=self.device),
            torch.tensor(neighbor_types, device=self.device),
            torch.tensor(neighbor_vectors, dtype=self.dtype, device=self.device),
            0  # Placeholder, usually reserved for batch index or flag
        ]

        # Model inference
        energy, energies, forces, stress, stresses = model(*input_tensors)
        self.results["energy"] = energy.cpu().detach().numpy().item()
        self.results["energies"] = energies.cpu().detach().numpy().reshape(-1)
        self.results["forces"] = forces.cpu().detach().numpy().reshape(-1, 3)
        if "stress" in properties:
            stress = stress.cpu().detach().numpy().reshape(3, 3)
            if any(atoms.get_pbc()):
                stress = -0.5 * (stress * stress.T) / atoms.get_volume()
                self.results["stress"] = stress

        if "stresses" in properties:
            stresses = stresses.cpu().detach().numpy().reshape(-1, 3, 3)
            if any(atoms.get_pbc()):
                self.results["stresses"] = -1.0 * stresses / atoms.get_volume()
            else:
                raise PropertyNotImplementedError




from typing import Dict
import torch
import numpy as np
from ase import Atoms
from abfml import to_graph
from abfml import ReadData
from abfml import FieldModel
import torch.nn as nn
from torch_geometric.data import Data
from torch_geometric.nn import GINEConv



class TestModel(FieldModel):
    def __init__(self):
        super().__init__(type_map=[0], cutoff=6.0, neighbor=100)

        # Edge encoder: MLP to process edge features
        self.edge_encoder = nn.Sequential(
            nn.Linear(1, 16),
            nn.ReLU(),
            nn.Linear(16, 32)
        )

        # GINEConv layer with MLP edge encoder
        self.conv1 = GINEConv(
            nn.Sequential(
                nn.Linear(1, 16),
                nn.ReLU(),
                nn.Linear(16, 1)
            ), edge_dim=32
        )

    def field(self,
              element_map: torch.Tensor,
              central_atoms: torch.Tensor,
              neighbor_indices: torch.Tensor,
              neighbor_types: torch.Tensor,
              neighbor_vectors: torch.Tensor,
              n_ghost: int) -> Dict[str, torch.Tensor]:
        # Convert atom data to graph representation
        graph = to_graph(central_atoms, neighbor_indices, neighbor_types, neighbor_vectors)
        x = graph[0]['x']
        edge_index = graph[0]['edge_index']
        edge_attr = graph[0]['relative_pos']

        # Process edge attributes
        edge_attr = edge_attr[:, 0:1] + torch.norm(edge_attr[:, 0:], dim=-1, keepdim=True)
        edge_attr = self.edge_encoder(edge_attr)

        # Create Data object for PyTorch Geometric
        data = Data(x=x, edge_index=edge_index, edge_attr=edge_attr)

        # Apply GINEConv
        x = self.conv1(x, edge_index, edge_attr)

        # Energy computations
        Ei = x.unsqueeze(0)
        Etot = torch.sum(Ei, dim=1)

        return {'Etot': Etot, 'Ei': Ei}

def main():
    # 设置原子的位置
    positions = [
        (5.0, 5.0, 5.0),  # 第一个原子的位置
        (5.0, 6.0, 6.0)  # 第二个原子的位置
    ]

    # 设置原子类型
    symbols = ['Cu', 'Cu']

    # 设置模拟盒子，较大的盒子
    cell = np.array([[30.0, 0.0, 0.0],
                     [0.0, 30.0, 0.0],
                     [0.0, 0.0, 30.0]])

    # 创建 Atoms 对象
    atoms = Atoms(symbols=symbols, positions=positions, cell=cell, pbc=True)
    # 设置 cutoff 和 max_neighbors
    cutoff = 5.0
    neighbor_dict = 100 # Si:4, O:3
    # 调用 calculate_neighbor
    element_types, central_atoms, neighbor_indices, neighbor_types, neighbor_vectors = ReadData.calculate_neighbor(
        atoms=atoms,
        cutoff=cutoff,
        neighbor=neighbor_dict
    )
    device = torch.device("cpu")
    element_types = torch.tensor(element_types, dtype=torch.long, device=device)
    central_atoms = torch.tensor(central_atoms, dtype=torch.long, device=device)
    neighbor_indices = torch.tensor(neighbor_indices, dtype=torch.long, device=device)
    neighbor_types = torch.tensor(neighbor_types, dtype=torch.long, device=device)
    neighbor_vectors = torch.tensor(neighbor_vectors, dtype=torch.float, device=device)

    model = TestModel()
    output = model(element_types, central_atoms, neighbor_indices, neighbor_types, neighbor_vectors, 0)
    print("Etot:", output[0])
    print("Ei:", output[1])
    print("force:", output[2])
    print("virial:", output[3])
    print("atomic_virial:", output[4])
    try:
        jit_model = torch.jit.script(model)
        jit_model.save('test-jit-script-model.pt')
        print("Model successfully scripted and saved.")
    except Exception as e:
        print(f"JIT model not compiled using script. Error: {e}")

if __name__ == "__main__":
    main()
    print("successful")



import torch
from abfml.param.param import TrainSet


def optim_init(model, param: TrainSet):
    """
    Initialize optimizer based on param.optimizer and param.lr.
    Optionally uses param.weight_decay and param.momentum if provided.
    """
    optim_name = param.optimizer
    lr = 0.001
    weight_decay = getattr(param, "weight_decay", 0.0)
    momentum = getattr(param, "momentum", 0.9)

    if optim_name == "Adam":
        optimizer = torch.optim.Adam(model.parameters(), lr=lr, weight_decay=weight_decay, eps=1e-5)
    elif optim_name == "Adamw":
        optimizer = torch.optim.AdamW(model.parameters(), lr=lr, weight_decay=weight_decay)
    elif optim_name == "sgd":
        optimizer = torch.optim.SGD(model.parameters(), lr=lr, momentum=momentum, weight_decay=weight_decay)
    elif optim_name == "rmsprop":
        optimizer = torch.optim.RMSprop(model.parameters(), lr=lr, momentum=momentum, weight_decay=weight_decay)
    elif optim_name == "adagrad":
        optimizer = torch.optim.Adagrad(model.parameters(), lr=lr, weight_decay=weight_decay)
    elif optim_name == "lbfgs":
        optimizer = torch.optim.LBFGS(model.parameters(), lr=lr, max_iter=20)
    elif optim_name == "user_defined":
        optimizer = None
    else:
        raise ValueError(f"Unsupported optimizer type: {param.optimizer}")

    return optimizer






import torch
from jsonschema import Draft7Validator, validators
from dataclasses import make_dataclass
from typing import Union

global_set_schema = {
    "type": "object",
    "properties": {
                    "type_map": {
                                "type": "array",
                                "items": {"type": "integer"}
                    },
                    "neighbor": {
                                "type": ["integer", "array"],
                                "items": {"type": "integer"}
                    },
                    "cutoff": {
                                "type": "number"
                    },
                    "dtype": {
                                "type": "string",
                                "enum": ["float32", "float64"],
                                "default": "float32"
                    },
                    "logger_file": {
                                "type": "string",
                                "default": "mlff.log"
                    }
    },
    "required": ["type_map", "neighbor", "cutoff"]
}

data_set_schema = {
    "type": "object",
    "properties": {
                    "file_format": {
                                "type": "string",
                                "default": None
                    },
                    "ratio": {
                                "type": "number", "minimum": 0, "maximum": 1,
                                "default": None},
                    "train_file":  {
                                "type": "array",
                                "items": {"type": "string"}
                    },
                    "valid_file": {
                                "type": "array",
                                "items": {"type": "string"},
                                "default": None
                    },
                    "shuffle": {
                                "type": "boolean",
                                "default": True
                    }
    },
    "required": ["train_file"]
}

train_set_schema = {
    "type": "object",
    "properties": {
                    "epoch": {
                                "type": "integer"
                    },
                    "optimizer": {
                                "type": "string",
                                "enum": ["Adam", "AdamW", "SGD"],
                                "default": "Adam"
                            },
                    "optim_set": {
                                "type": "object",
                                "additionalProperties": True
                    },

                    "freeze_param": {
                                "type": "array",
                                "items": {"type": "string"},
                                "default": None
                    },
                    "save_check": {
                                "type": "integer",
                                "default": 5
                    },
                    "batch_size": {
                                "type": "integer",
                                "default": 1
                    },
                    "print_freq": {
                                "type": "integer",
                                "default": 1
                    },
                    "model_file": {
                                "type": "string",
                                "default": "model.pt"
                    }
    },
    "if": {
        "properties": {"optimizer": {"const": "Adam"}}
    },
    "then": {
        "properties": {"optim_set": {
                            "properties": {"AdamW": {"type": "object"}}}}
    },
    "else": {
        "if": {
            "properties": {"optimizer": {"const": "SGD"}}
        },
        "then": {
            "properties": {"optim_set": {
                            "properties": {"SGD": {"type": "object"}}}}
        },
        "else": {
            "properties": {"optim_set": {
                            "properties": {"RMSProp": {"type": "object"}}}}
        }
    },
    "required": ["epoch", "optimizer"]
}

loss_set_schema = {
    "type": "object",
    "properties": {
                    "start_energy_weight": {
                                "type": "number",
                                "default": 0
                    },
                    "limit_energy_weight": {
                                "type": "number",
                                "default": 0
                    },
                    "start_force_weight": {
                                "type": "number",
                                "default": 0
                    },
                    "limit_force_weight": {
                                "type": "number",
                                "default": 0
                    },
                    "start_virial_weight": {
                                "type": "number",
                                "default": 0
                    },
                    "limit_virial_weight": {
                                "type": "number",
                                "default": 0
                    },
                    "start_ei_weight": {
                                "type": "number",
                                "default": 0
                    },
                    "limit_ei_weight": {
                                "type": "number",
                                "default": 0
                    },
                    "loss_file": {
                                "type": "string",
                                "default": "loss.curve"
                    }
    }
}

learning_rate_schema = {
                "type": "object",
                "properties": {
                    "type_name": {
                        "type": "string",
                        "enum": ["exp"],
                        "default": "exp"
                    },
                    "decay_step": {
                        "type": "integer",
                        "default": 1000
                    },
                    "decay_rate": {
                        "type": "number",
                        "default": None
                    },
                    "start_lr": {
                        "type": "number",
                        "default": 1e-3
                    },
                    "limit_lr": {
                        "type": "number",
                        "default": 1e-8
                    },
                }
}

fitting_net_schema= {
            "type": "object",
            "properties": {
                        "network_size": {
                            "type": "array",
                            "items": {"type": "integer"}
                        },
                        "bias": {
                            "type": "boolean",
                            "default": True
                        },
                        "resnet_dt": {
                            "type": "boolean",
                            "default": False
                        },
                        "activate_function": {
                            "type": "string",
                            "default": "tanh",
                            "enum": ["tanh", "relu", "sigmoid", "softplus", "softsign"]
                        }
            },
            "required": ["network_size"]
}

deep_se_e = {
          "type": "object",
          "properties": {
                        "R_max": {
                            "type": "number"
                        },
                        "R_min": {
                            "type": "number"
                        },
                        "smooth_fun": {
                            "type": "string",
                            "default": "poly2_r",
                            "enum": ["cos", "cos_r", "tanh_u", "exp", "poly1_r", "poly2_r", "poly3_r"]
                        },
                        "network_size": {
                            "type": "array",
                            "items": {"type": "integer"}
                        },
                        "bias": {
                            "type": "boolean",
                            "default": True
                        },
                        "resnet_dt": {
                            "type": "boolean",
                            "default": False
                        },
                        "activate_function": {
                            "type": "string",
                            "default": "tanh",
                            "enum": ["tanh", "relu", "sigmoid", "softplus", "softsign"]
                        },
                        "M2": {
                            "type": "integer",
                            "default": 16
                        }
          },
          "required": ["R_max", "R_min", "smooth_fun", "network_size"]
}


deep_se_e2_e = {
          "type": "object",
          "properties": {
                        "type_name": {
                            "type": "string",
                            "const": "se_e2_e"
                        },
                        "R_max": {
                            "type": "number"
                        },
                        "R_min": {
                            "type": "number"
                        },
                        "smooth_fun": {
                            "type": "string",
                            "default": "poly2_r",
                            "enum": ["cos", "cos_r", "tanh_u", "exp", "poly1_r",  "poly2_r","poly3_r"]
                        },
                        "network_size": {
                            "type": "array",
                            "items": {"type": "integer"}
                        },
                        "bias": {
                            "type": "boolean",
                            "default": True
                        },
                        "resnet_dt": {
                            "type": "boolean",
                            "default": False
                        },
                        "activate_function": {
                            "type": "string",
                            "default": "tanh",
                            "enum": ["tanh", "relu", "sigmoid", "softplus", "softsign"]
                        },
                        "M2": {
                            "type": "integer",
                            "default": 16
                        },
                        "A": {
                            "type": ["null", "string"]
                        }
          },
          "required": ["type_name", "R_max", "R_min", "smooth_fun", "network_size"]
}

BP_param_schema = {
                   "SymFunRad": {
                                   "type": "object",
                                   "properties": {
                                       "type_name": {
                                                   "type": "string",
                                                   "const": "SymFunRad"
                                       },
                                       "R_max": {
                                                   "type": "number"
                                       },
                                       "R_min": {
                                                   "type": "number"
                                       },
                                       "num_basis": {
                                                   "type": "integer"
                                       },
                                       "smooth_fun": {
                                                   "type": "string",
                                                   "default": "poly2_r",
                                                   "enum": ["cos", "cos_r", "tanh_u", "exp",
                                                            "poly1_r", "poly2_r", "poly3_r"]
                                       },
                                       "eta": {
                                           "oneOf": [
                                               {
                                                   "type": "number"
                                               },
                                               {
                                                   "type": "array",
                                                   "items": {
                                                       "type": "number"
                                                   }
                                               }
                                           ]
                                       },
                                       "rs": {
                                           "oneOf": [
                                               {
                                                   "type": "number"
                                               },
                                               {
                                                   "type": "array",
                                                   "items": {
                                                       "type": "number"
                                                   }
                                               }
                                           ]
                                       },
                                       "rs_fun": {
                                           "type": "string",
                                           "default": "linear",
                                           "enum": ["linear"]
                                       }
                                   },
                                   "required": ["R_max", "R_min", "smooth_fun"],
                                   "oneOf": [
                                       {
                                           "properties": {
                                               "num_basis": {
                                                   "type": "integer"
                                               },
                                               "rs_fun": {
                                                   "type": "string",
                                                   "enum": ["linear"]
                                               }
                                           },
                                           "required": ["num_basis", "rs_fun"]
                                       },
                                       {
                                           "properties": {
                                               "eta": {
                                                   "type": "array",
                                                   "items": {
                                                       "type": "number"
                                                   }
                                               },
                                               "rs": {
                                                   "type": "array",
                                                   "items": {
                                                       "type": "number"
                                                   }
                                               }
                                           },
                                           "required": ["eta", "rs"]
                                       }
                                   ]

                   },
                   "SymFunAng": {
                           "type": "object",
                           "properties": {
                                       "type_name": {
                                            "type": "string"
                                       },
                                       "R_max": {
                                           "type": "number"
                                       },
                                       "R_min": {
                                           "type": "number"
                                       },
                                       "num_basis": {
                                           "type": "integer"
                                       },
                                       "smooth_fun": {
                                           "type": "string",
                                           "default": "poly2_r",
                                           "enum": ["cos", "cos_r", "tanh_u", "exp", "poly1_r", "poly2_r", "poly3_r"]
                                       },
                                       "eta": {
                                           "type": ["number", "array"],
                                           "items": {
                                                    "type": "number"
                                           }
                                       },
                                       "lambd": {
                                           "type": ["number", "array"],
                                           "items": {
                                                    "type": "number"
                                           }
                                       },
                                       "zeta": {
                                           "type": ["number", "array"],
                                           "items": {
                                                    "type": "number"
                                           }
                                       },
                                       "rs_fun": {
                                           "type": "string"
                                       }
                           },
                           "required": ["R_max", "R_min", "num_basis", "eta", "lambd", "zeta", "rs_fun"]
                   }
}


NEP_schema = {
          "type": "object",
          "properties": {
                        "R_max_R": {
                            "type": "number",
                            "default": 6.0
                        },
                        "R_max_A": {
                            "type": "number",
                            "default": 6.0
                        },
                        "n_max_R": {
                            "type": "integer",
                            "default": 4
                        },
                        "n_max_A": {
                            "type": "integer",
                            "default": 4
                        },
                        "k_basis_R": {
                            "type": "integer",
                            "default": 4
                        },
                        "k_basis_A": {
                            "type": "integer",
                            "default": 4
                        },
                        "l_3b_max": {
                            "type": "integer",
                            "default": 4,
                            "enum": [4]
                        },
                        "l_4b_max": {
                            "type": "integer",
                            "default": 2,
                            "enum": [0, 2]
                        },
                        "l_5b_max": {
                            "type": "integer",
                            "default": 0,
                            "enum": [0, 1]
                        },
                        "smooth_fun": {
                            "type": "string",
                            "default": "cos",
                            "enum": ["cos", "cos_r", "tanh_u", "exp", "poly1_r",  "poly2_r","poly3_r"]
                        },
          }
}


# Define the validator, including default values
def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for error in validate_properties(validator, properties, instance, schema):
            yield error

        for property_value, sub_schema in properties.items():
            if "default" in sub_schema:
                instance.setdefault(property_value, sub_schema["default"])

    return validators.extend(
        validator_class, {"properties": set_defaults},
    )


# Creating Validators with Default Values
DefaultValidatingDraft7Validator = extend_with_default(Draft7Validator)


# Dynamically generated parameter classes
def create_dataclass(schema: dict):
    properties = schema.get('properties', {})
    fields = []
    type_dict = {
                "string": str,
                "integer": int,
                "number": float,
                "boolean": bool,
                "array": list
    }
    for key, value in properties.items():
        type_class = type_dict.get("type", Union)
        default_value = value.get('default', None)
        fields.append((key, type_class, default_value))
    return make_dataclass(cls_name='SchemaClass', fields=fields)


# Generate parameter classes
GlobalSet = create_dataclass(global_set_schema)
DataSet = create_dataclass(data_set_schema)
TrainSet = create_dataclass(train_set_schema)
LossSet = create_dataclass(loss_set_schema)
LrSet = create_dataclass(learning_rate_schema)


class DeepSe:
    def __init__(self, fitting_config: dict, embedding_config: dict):
        validator = DefaultValidatingDraft7Validator(fitting_net_schema)
        validator.validate(fitting_config)

        validator = DefaultValidatingDraft7Validator(deep_se_e)
        validator.validate(embedding_config)

        self.fitting_config = fitting_config
        self.embedding_config = embedding_config


class BPDescriptor:
    def __init__(self, fitting_config: dict, feature_config: dict):
        self.fitting_config = fitting_config
        self.bp_features_information = []
        self.bp_features_param = []
        for i, feature_information in enumerate(feature_config['list']):
            feature_tensor = {}
            feature_str = {}
            if feature_information.get('type_name') == 'SymFunRad':
                validator = DefaultValidatingDraft7Validator(BP_param_schema['SymFunRad'])
                validator.validate(feature_information)
                feature_name = 'SymFunRad'
                feature_tensor['R_min'] = torch.Tensor([feature_information['R_min']])
                feature_tensor['R_max'] = torch.Tensor([feature_information['R_max']])
                if feature_information.get('rs_fun', None) is not None:
                    feature_tensor['rs'] = BPDescriptor.rs_fun(feature_information['R_min'],
                                                               feature_information['R_max'],
                                                               feature_information['num_basis'],
                                                               feature_information['rs_fun'])
                else:
                    feature_tensor['rs'] = torch.Tensor([feature_information['rs']])
                num_basis = feature_tensor['rs'].shape[0]
                if isinstance(feature_information['eta'], float):
                    feature_tensor['eta'] = (torch.Tensor([feature_information['eta']])).repeat(num_basis)
                else:
                    feature_tensor['eta'] = torch.Tensor([feature_information['eta']])
                feature_str['smooth_fun'] = feature_information['smooth_fun']
            elif 'SymFunAng' in feature_information.get('type_name'):
                validator = DefaultValidatingDraft7Validator(BP_param_schema['SymFunAng'])
                validator.validate(feature_information)
                feature_name = feature_information.get('type_name')
                feature_tensor['R_min'] = torch.Tensor([feature_information['R_min']])
                feature_tensor['R_max'] = torch.Tensor([feature_information['R_max']])
                if feature_information.get('rs_fun', None) is not None:
                    feature_tensor['rs'] = BPDescriptor.rs_fun(feature_information['R_min'],
                                                               feature_information['R_max'],
                                                               feature_information['num_basis'],
                                                               feature_information['rs_fun'])
                else:
                    feature_tensor['rs'] = torch.Tensor([feature_information['rs']])
                num_basis = feature_tensor['rs'].shape[0]
                if isinstance(feature_information['eta'], float):
                    feature_tensor['eta'] = (torch.Tensor([feature_information['eta']])).repeat(num_basis)
                    feature_tensor['lambd'] = (torch.Tensor([feature_information['lambd']])).repeat(num_basis)
                    feature_tensor['zeta'] = (torch.Tensor([feature_information['zeta']])).repeat(num_basis)
                else:
                    feature_tensor['eta'] = torch.Tensor([feature_information['eta']])
                    feature_tensor['lambd'] = torch.Tensor([feature_information['lambd']])
                    feature_tensor['zeta'] = torch.Tensor([feature_information['zeta']])
                feature_str['smooth_fun'] = feature_information['smooth_fun']
            else:
                raise ValueError('Wrong type of BP descriptor')
            self.bp_features_information.append(tuple([feature_name, num_basis]))
            self.bp_features_param.append(tuple([feature_tensor, feature_str]))

    @staticmethod
    def rs_fun(R_min: float, R_max: float, num_basis: int, type_name: str) -> torch.Tensor:
        if type_name == 'linear':
            rs = torch.linspace(R_min, R_max, num_basis)
        else:
            raise ValueError('Wrong type of rs_fun')
        return rs


class NEPParam:
    def __init__(self, fitting_config: dict, feature_config: dict):
        validator = DefaultValidatingDraft7Validator(fitting_net_schema)
        validator.validate(fitting_config)

        validator = DefaultValidatingDraft7Validator(NEP_schema)
        validator.validate(feature_config)

        self.fitting_config = fitting_config
        self.feature_config = feature_config


class Param:
    def __init__(self, input_dict: dict):
        self.input_dict_new = {}
        global_dict = input_dict['global_setting']
        validator = DefaultValidatingDraft7Validator(global_set_schema)
        validator.validate(global_dict)
        self.GlobalSet = GlobalSet(**global_dict)
        if isinstance(self.GlobalSet.neighbor, list):
            self.GlobalSet.neighbor = {t: n for t, n in zip(self.GlobalSet.type_map, self.GlobalSet.neighbor)}

        data_dict = input_dict['data_setting']
        validator = DefaultValidatingDraft7Validator(data_set_schema)
        validator.validate(data_dict)
        self.DataSet = DataSet(**data_dict)

        train_dict = input_dict['train_setting']
        validator = DefaultValidatingDraft7Validator(train_set_schema)
        validator.validate(train_dict)
        self.TrainSet = TrainSet(**train_dict)

        loss_dict = input_dict['loss_setting']
        validator = DefaultValidatingDraft7Validator(loss_set_schema)
        validator.validate(loss_dict)
        self.LossSet = LossSet(**loss_dict)

        learning_rate_dict = input_dict['learning_rate']
        validator = DefaultValidatingDraft7Validator(learning_rate_schema)
        validator.validate(learning_rate_dict)
        self.LrSet = LrSet(**learning_rate_dict)

        self.model_name = input_dict['model_setting']['name']
        if input_dict['model_setting']['name'] in ['dp_se_e2_a', 'dp_se_e2_r', 'dp_se_e3']:
            fitting_config = input_dict['model_setting']['fitting_net']
            embedding_config = input_dict['model_setting']['descriptor']
            self.DeepSe = DeepSe(fitting_config=fitting_config, embedding_config=embedding_config)
        elif input_dict['model_setting']['name'] == 'BPMlp':
            fitting_config = input_dict['model_setting']['fitting_net']
            feature_config = input_dict['model_setting']['descriptor']
            self.BPDescriptor = BPDescriptor(fitting_config=fitting_config, feature_config=feature_config)
        elif input_dict['model_setting']['name'] == "NEP":
            fitting_config = input_dict['model_setting']['fitting_net']
            feature_config = input_dict['model_setting']['descriptor']
            self.NEPParam = NEPParam(fitting_config=fitting_config, feature_config=feature_config)
        elif input_dict['model_setting']['name'] == "test":
            self.test = self.Test(input_dict["model_setting"])
        elif input_dict['model_setting']['name'] == "user_defined":
            self.user_defined = self.UserDefined(input_dict["model_setting"])

        else:
            raise Exception(f"Undefined keyword: , Please read the manual!")

    class Test:
        def __init__(self, input_dict):
            self.input_dict = input_dict

    class UserDefined:
        config: dict = {}
        model_path: str = ""
        field_name: str = ""
        normal_name: str = None

        def __init__(self, config: dict):
            if 'descriptor' in config.keys():
                self.config = config["descriptor"]
            else:
                raise Exception("Please write the custom parameters in the descriptor")

            if "model_path" in config.keys():
                self.model_path = config["model_path"]
            else:
                raise Exception("Undefined model path:{model_path} ?")

            self.field_name = config["field_name"]

            if "normal_name" in config.keys():
                self.normal_name = config["normal_name"]
            else:
                self.normal_name = None





import torch
from typing import List, Union
from abfml.param.param import Param
from abfml.core.model.method import NormalModel


class DPNormal(NormalModel):
    def __init__(self,
                 normal_data,
                 param_class: Param,
                 normal_rate: Union[float, str],
                 coordinate_matrix: str):
        super().__init__(normal_data=normal_data,
                         param_class=param_class,
                         normal_rate=normal_rate,
                         is_get_energy_shift=True)
        self.coordinate_matrix = coordinate_matrix
        self.R_min = None
        self.R_max = None
        self.smooth_type = None

        self.reset(param_class=param_class)
        self.initialize()

    def reset(self, param_class: Param):
        self.R_min = param_class.DeepSe.embedding_config['R_min']
        self.R_max = param_class.DeepSe.embedding_config['R_max']
        self.smooth_type = param_class.DeepSe.embedding_config['smooth_fun']

    def normal(self, normal_loader, param_class: Param):
        type_map = param_class.GlobalSet.type_map
        ntype = len(type_map)
        type_num = torch.zeros(ntype)
        std_mean: List[torch.Tensor] = [torch.zeros(ntype, 4, requires_grad=False),
                                        torch.zeros(ntype, 4, requires_grad=False)]
        if self.coordinate_matrix == 'a':
            from abfml.core.model.dpse import DpSe2a
            calculate_coordinate_matrix = DpSe2a.calculate_coordinate_matrix
        elif self.coordinate_matrix == 'r':
            from abfml.core.model.dpse import DpSe2r
            calculate_coordinate_matrix = DpSe2r.calculate_coordinate_matrix
        else:
            raise KeyError("Key word of coordinate_matrix have 'a' or 'r'")

        for i, image_batch in enumerate(normal_loader):
            neighbor_vectors = image_batch["neighbor_vectors"]
            central_atoms = image_batch["central_atoms"]
            element_types = image_batch["element_types"][0].to(torch.int64).tolist()

            Ri = calculate_coordinate_matrix(R_min=self.R_min,
                                             R_max=self.R_max,
                                             smooth_type=self.smooth_type,
                                             Rij=neighbor_vectors)
            # I think the sample standard deviation should be used instead of the standard deviation
            for i_type, element in enumerate(element_types):
                mask = (central_atoms == element)
                s_rij = Ri[:, :, :, 0][mask]
                s_rij_std, s_rij_mean = torch.std_mean(s_rij)

                if Ri.shape[-1] == 4:
                    xyz = Ri[:, :, :, 1:][mask]
                    xyz_std, xyz_mean = torch.std_mean(xyz)
                else:
                    xyz_std, xyz_mean = 0.0, 0.0

                indices = type_map.index(element)
                type_num[indices] += 1

                std_mean[0][indices][0] = std_mean[0][indices][0] + s_rij_std
                std_mean[0][indices][1] = std_mean[0][indices][1] + xyz_std
                std_mean[0][indices][2] = std_mean[0][indices][2] + xyz_std
                std_mean[0][indices][3] = std_mean[0][indices][3] + xyz_std
                std_mean[1][indices][0] = std_mean[1][indices][0] + s_rij_mean

        type_num[type_num == 0] = 1
        for i in range(ntype):
            std_mean[0][i] = std_mean[0][i] / type_num[i]
            std_mean[1][i] = std_mean[1][i] / type_num[i]

        return std_mean


class FeatureNormal(NormalModel):
    def __init__(self,
                 normal_data,
                 param_class: Param,
                 normal_rate: Union[float, str],
                 feature_name: str):
        super().__init__(normal_data=normal_data,
                         param_class=param_class,
                         normal_rate=normal_rate,
                         is_get_energy_shift=True)
        self.feature_name = feature_name
        self.initialize()

    def normal(self, normal_loader, param_class: Param):
        type_map = param_class.GlobalSet.type_map
        ntype = len(type_map)
        type_num = torch.zeros(ntype)
        std_mean: List[torch.Tensor] = [torch.zeros(ntype,  requires_grad=False),
                                        torch.zeros(ntype, requires_grad=False)]
        from abfml.core.model.bpmlp import BPMlp
        for i, image_batch in enumerate(normal_loader):
            neighbor_vectors = image_batch["neighbor_vectors"]
            element_types = image_batch["element_types"][0]
            central_atoms = image_batch["central_atoms"]
            neighbor_indices = image_batch["neighbor_indices"]
            neighbor_types = image_batch["neighbor_types"]

            element_type = image_batch["element_type"][0].to(torch.int64)
            Gi = BPMlp.calculate_bp_feature(type_map=type_map,
                                            bp_features_information=param_class.BPDescriptor.bp_features_information,
                                            bp_features_param=param_class.BPDescriptor.bp_features_param,
                                            element_map=element_type,
                                            neighbor_indices=neighbor_indices,
                                            neighbor_types=neighbor_types,
                                            neighbor_vectors=neighbor_vectors)

            # I think the sample standard deviation should be used instead of the standard deviation
            Gi = Gi.clone().detach()
            for i_type, element in enumerate(element_type):
                mask = (central_atoms == element)
                feature = Gi[mask]
                feature_std, feature_mean = torch.std_mean(feature)

                indices = type_map.index(element)
                type_num[indices] += 1

                std_mean[0][indices] = std_mean[0][indices] + feature_std
                std_mean[1][indices] = std_mean[1][indices] + feature_mean

        type_num[type_num == 0] = 1
        for i in range(ntype):
            std_mean[0][i] = std_mean[0][i] / type_num[i]
            std_mean[1][i] = std_mean[1][i] / type_num[i]

        return std_mean




import sys
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde


# Set font style to Times New Roman
plt.rcParams.update({'font.family': 'serif', 'font.serif': ['Times New Roman']})
plt.rcParams['mathtext.default'] = 'regular'

def plot_loss_curve(file_name):
    """
    Reads a loss file and plots the training and validation RMSE curves.

    Parameters:
        file_name (str): Path to the input CSV file.
    """
    # Read the data file
    data_pd = pd.read_csv(file_name, sep=r'\s+', header=0)

    # List of RMSE metrics to be plotted
    label_list = ["Loss", "E_tot", "Force", "Ei", "Virial"]

    for label in label_list:
        train_key = f"T_RMSE_{label}"
        valid_key = f"V_RMSE_{label}"

        # Check if both training and validation keys exist in the dataset
        if train_key in data_pd and valid_key in data_pd:
            # Create a figure and axis with a larger size for better readability
            fig, ax = plt.subplots(figsize=(12, 8))

            # Plot training and validation RMSE curves
            ax.plot(data_pd["Epoch"], data_pd[train_key], label="Train", linestyle="-", marker="o",
                    markersize=4)
            ax.plot(data_pd["Epoch"], data_pd[valid_key], label="Valid", linestyle="--", marker="s",
                    markersize=4)

            # Set title and axis labels
            ax.set_title(f"{label} RMSE vs Epoch", fontsize=18)
            ax.set_xlabel("Epoch", fontsize=18)
            ax.set_ylabel(f"{label} RMSE", fontsize=18)
            ax.tick_params(axis='both', labelsize=16)

            # Add grid lines with transparency
            ax.grid(linestyle="--", alpha=0.6)

            # Improve legend visibility
            ax.legend(fontsize=18, loc="best", frameon=True, edgecolor="gray")

            # Save the figure as a high-resolution PNG file
            plt.savefig(f'loss_{label}.png', dpi=300, bbox_inches="tight")
            plt.close()


def plot_scatter(dft, predict, predict_key, rmse, r2, unit):
    """
    Plot scatter diagram comparing DFT and predicted values.
    """
    plt.figure(figsize=(10, 8))  # Increase image size
    plt.scatter(dft, predict, s=20, label='Data', alpha=0.6)  # Enhance scatter plot effect

    # Diagonal line (y = x, ideal fit line)
    data_min = min(dft.min(), predict.min())
    data_max = max(dft.max(), predict.max())
    plt.plot([data_min, data_max], [data_min, data_max], color='black', linestyle='--', linewidth=1.5, label='$y=x$')

    # Annotate RMSE and R² values
    plt.text(0.05, 0.95, f"RMSE: {rmse:.4f}{unit}\nR²: {r2:.4f}", transform=plt.gca().transAxes,
             fontsize=18, verticalalignment='top', bbox=dict(facecolor='white', alpha=0.8))

    # Set title, axis labels, and grid
    plt.title(f'{predict_key}: DFT vs Predict', fontsize=22)
    plt.xlabel(f'{predict_key}$_{{DFT}}$', fontsize=20)
    plt.ylabel(f'{predict_key}$_{{Predict}}$', fontsize=20)
    plt.xlim(data_min, data_max)
    plt.ylim(data_min, data_max)
    plt.tick_params(axis='both', labelsize=18)
    plt.grid(True, linestyle='--', alpha=0.5)
    plt.legend()

    # Save scatter plot
    plt.savefig(f'{predict_key}_scatter.png', dpi=300, bbox_inches='tight')
    plt.close()


def plot_error_distribution(difference, predict_key, rmse, r2, unit):
    """
    Plot error distribution using histogram and KDE.
    """
    plt.figure(figsize=(10, 8))

    # Histogram part
    counts, bins, _ = plt.hist(difference, bins=40, density=True, alpha=0.7, color='#1f77b4', edgecolor='black')

    # KDE (Kernel Density Estimation) smoothed density curve
    kde = gaussian_kde(difference)
    kde_x = np.linspace(bins.min(), bins.max(), 300)
    plt.plot(kde_x, kde(kde_x), color='red', linewidth=2, label='Density')

    # Annotate RMSE and R² values
    plt.text(0.65, 0.95, f"RMSE: {rmse:.4f}{unit}\nR²: {r2:.4f}", transform=plt.gca().transAxes,
             fontsize=18, verticalalignment='top', bbox=dict(facecolor='white', alpha=0.8))

    # Set title, axis labels, and grid
    plt.title(f'{predict_key} Error Distribution', fontsize=22)
    plt.xlabel('Error (DFT - Predict)', fontsize=20)
    plt.ylabel('Density', fontsize=20)
    plt.tick_params(axis='both', labelsize=18)
    plt.grid(True, linestyle='--', alpha=0.5)
    plt.legend()

    # Save error distribution plot
    plt.savefig(f'ErrorDistributionOf_{predict_key}.png', dpi=300, bbox_inches='tight')
    plt.close()

if __name__ == '__main__':
    # Ensure the script is run with a valid input file
    if len(sys.argv) < 2:
        print("Usage: python script.py <loss_file>")
        sys.exit(1)

    # Get the loss file from command-line arguments
    loss_file = sys.argv[1]
    plot_loss_curve(loss_file)



import argparse
import textwrap
from typing import List, Optional


class RawTextArgumentDefaultsHelpFormatter(
    argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter
):
    """This formatter is used to print multile-line help message with default value."""


def main_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="ABFML: A package for rapid building, fitting, and application of machine learning force fields",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers = parser.add_subparsers(title="ABFML-kit subcommands", dest="command")

    # **************************************   train parser start **************************************
    parser_train = subparsers.add_parser(
        name="train",
        parents=[],
        help="train a model",
        formatter_class=RawTextArgumentDefaultsHelpFormatter,
        epilog=textwrap.dedent(
            """\
        examples:
            abfml train input.json
            abfml train input.json --restart model.ckpt
            abfml train input.json --init-model model.ckpt
        """
        ),
    )
    parser_train.add_argument(
        "INPUT", help="the input parameter file in json format."
    )
    parser_train_subgroup = parser_train.add_mutually_exclusive_group()
    parser_train_subgroup.add_argument(
        "-i",
        "--init_model",
        type=str,
        default=None,
        help="Initialize the model by the provided checkpoint.",
    )
    parser_train_subgroup.add_argument(
        "-r",
        "--restart",
        type=str,
        default=None,
        help="Initialize the training from the frozen model.",
    )
    # **************************************   train parser end   **************************************

    # **************************************   valid parser start **************************************
    parser_valid = subparsers.add_parser(
        name="valid",
        parents=[],
        help="valid the model",
        formatter_class=RawTextArgumentDefaultsHelpFormatter,
        epilog=textwrap.dedent(
            """\
        examples:
            abfml valid -m model.pt -f /path/to/system -n 30
        """
        ),
    )
    parser_valid.add_argument(
        "-m",
        "--model",
        default="model.pt",
        type=str,
        help="Valid model file to import",
    )
    parser_valid.add_argument(
        "-s",
        "--shuffle",
        default=False,
        action="store_true",
        help="Shuffle data and randomised",
    )
    parser_valid.add_argument(
        "-f",
        "--datafile",
        default=None,
        nargs="+",
        type=str,
        help="The path to file of test list.",
    )
    parser_valid.add_argument(
        "-p",
        "--plot",
        action="store_true",
        help="information",
    )
    parser_valid.add_argument(
        "-n", "--numb_test",
        default=100,
        type=int,
        help="The number of data for test"
    )

    # **************************************   valid parser end   **************************************

    # **************************************   check parser start **************************************
    parser_check = subparsers.add_parser(
        "check",
        parents=[],
        help="check the model",
        formatter_class=RawTextArgumentDefaultsHelpFormatter,
        epilog=textwrap.dedent(
            """\
        examples:
            abfml check -m model.pt 
            abfml check -i input.json -d float32
        """
        ),
    )

    parser_check_subgroup = parser_check.add_mutually_exclusive_group()
    parser_check_subgroup.add_argument(
        "-i",
        "--input",
        default=None,
        type=str,
        help="the input parameter file in json format.",
    )
    parser_check_subgroup.add_argument(
        "-m",
        "--model_name",
        default=None,
        type=str,
        help="check model file to import",
    )
    parser_check.add_argument(
        "-d",
        "--dtype",
        default="float32",
        type=str,
        help="Accuracy used when checking the model",
    )
    # ************************************** check parser end   **************************************

    return parser


def main_parse_args(args: Optional[List[str]] = None) -> argparse.Namespace:
    """Parse arguments and convert argument strings to objects.

    Parameters
    ----------
    args : List[str]
        list of command line arguments, main purpose is testing default option None
        takes arguments from sys.argv

    Returns
    -------
    argparse.Namespace
        the populated namespace
    """
    parser = main_parser()
    parsed_args = parser.parse_args(args=args)
    if parsed_args.command is None:
        parser.print_help()
    return parsed_args



import torch
import torch.nn as nn
from typing import List, Tuple
from abfml.core.model.math_fun import ActivationModule


class EmbeddingNet(nn.Module):
    def __init__(self,
                 network_size: List[int],
                 activate: str = 'tanh',
                 bias: bool = True,
                 resnet_dt: bool = False):
        super(EmbeddingNet, self).__init__()
        self.network_size = [1] + network_size  # [1, 25, 50, 100]
        self.bias = bias
        self.resnet_dt = resnet_dt
        self.activate = ActivationModule(activation_name=activate)
        self.linear = nn.ModuleList()
        self.resnet = nn.ParameterList()
        for i in range(len(self.network_size) - 1):
            self.linear.append(nn.Linear(in_features=self.network_size[i],
                                         out_features=self.network_size[i + 1], bias=self.bias))
            if self.bias:
                nn.init.normal_(self.linear[i].bias, mean=0.0, std=1.0)
            if self.network_size[i] == self.network_size[i+1] or self.network_size[i] * 2 == self.network_size[i+1]:
                resnet_tensor = torch.Tensor(1, self.network_size[i + 1])
                nn.init.normal_(resnet_tensor, mean=0.1, std=0.001)
                self.resnet.append(nn.Parameter(resnet_tensor, requires_grad=True))
            nn.init.normal_(self.linear[i].weight, mean=0.0,
                            std=(1.0 / (self.network_size[i] + self.network_size[i + 1]) ** 0.5))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        m = 0
        for i, linear in enumerate(self.linear):
            hidden = linear(x)
            hidden = self.activate(hidden)
            if self.network_size[i] == self.network_size[i+1] and self.resnet_dt:
                for ii, resnet in enumerate(self.resnet):
                    if ii == m:
                        x = hidden * resnet + x
                m = m + 1
            elif self.network_size[i] == self.network_size[i+1] and (not self.resnet_dt):
                x = hidden + x
            elif self.network_size[i] * 2 == self.network_size[i+1] and self.resnet_dt:
                for ii, resnet in enumerate(self.resnet):
                    if ii == m:
                        x = hidden * resnet + torch.cat((x, x), dim=-1)
                m = m + 1
            elif self.network_size[i] * 2 == self.network_size[i+1] and (not self.resnet_dt):
                x = hidden + torch.cat((x, x), dim=-1)
            else:
                x = hidden
        return x


class FittingNet(nn.Module):
    def __init__(self,
                 network_size: List[int],
                 activate: str,
                 bias: bool,
                 resnet_dt: bool,
                 energy_shift: float):
        super(FittingNet, self).__init__()
        self.network_size = network_size + [1]  # [input, 25, 50, 100, 1]
        self.bias = bias
        self.resnet_dt = resnet_dt
        self.activate = ActivationModule(activation_name=activate)
        self.linear = nn.ModuleList()
        self.resnet = nn.ParameterList()
        for i in range(len(self.network_size) - 1):
            if i == (len(self.network_size) - 2):
                self.linear.append(nn.Linear(in_features=self.network_size[i],
                                             out_features=self.network_size[i + 1], bias=True))
                nn.init.normal_(self.linear[i].bias, mean=energy_shift, std=1.0)
            else:
                self.linear.append(nn.Linear(in_features=self.network_size[i],
                                             out_features=self.network_size[i + 1], bias=self.bias))
                if self.bias:
                    nn.init.normal_(self.linear[i].bias, mean=0.0, std=1.0)
            if self.network_size[i] == self.network_size[i+1] and self.resnet_dt:
                resnet_tensor = torch.Tensor(1, self.network_size[i + 1])
                nn.init.normal_(resnet_tensor, mean=0.1, std=0.001)
                self.resnet.append(nn.Parameter(resnet_tensor, requires_grad=True))
            nn.init.normal_(self.linear[i].weight, mean=0.0,
                            std=(1.0 / (self.network_size[i] + self.network_size[i + 1]) ** 0.5))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        m = 0
        for i, linear in enumerate(self.linear):
            if i == (len(self.network_size) - 2):
                hidden = linear(x)
                x = hidden
            else:
                hidden = linear(x)
                hidden = self.activate(hidden)
                if self.network_size[i] == self.network_size[i+1] and self.resnet_dt:
                    for ii, resnet in enumerate(self.resnet):
                        if ii == m:
                            x = hidden * resnet + x
                    m = m + 1
                elif self.network_size[i] == self.network_size[i+1] and (not self.resnet_dt):
                    x = hidden + x
                elif self.network_size[i] * 2 == self.network_size[i+1] and self.resnet_dt:
                    for ii, resnet in enumerate(self.resnet):
                        if ii == m:
                            x = hidden * resnet + torch.cat((x, x), dim=-1)
                    m = m + 1
                elif self.network_size[i] * 2 == self.network_size[i+1] and (not self.resnet_dt):
                    x = hidden + torch.cat((x, x), dim=-1)
                else:
                    x = hidden
        return x


class Dense(nn.Module):
    def __init__(self, num_channels: int, in_features: int, out_features: int, bias: bool = True, activate: str = 'tanh', residual: bool = False) -> None:
        super().__init__()
        self.num_channels = num_channels
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.Tensor(num_channels, out_features, in_features))
        if bias:
            self.bias = nn.Parameter(torch.Tensor(num_channels, out_features))
        else:
            self.register_parameter('bias', None)
        self.activation = ActivationModule(activation_name=activate)
        self.residual = residual
        self.reset_parameters()

    def reset_parameters(self) -> None:
        for w in self.weight:
            nn.init.kaiming_uniform_(w, a=5**0.5)
        if self.bias is not None:
            for b, w in zip(self.bias, self.weight):
                fan_in, _ = nn.init._calculate_fan_in_and_fan_out(w)
                bound = 1 / (fan_in ** 0.5)
                nn.init.uniform_(b, -bound, bound)

    def forward(self, xx: Tuple[torch.Tensor, torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor]:
        x, channels = xx
        weight: torch.Tensor = self.weight[channels]
        output: torch.Tensor = torch.bmm(x.transpose(0, 1), weight.transpose(1, 2)).transpose(0, 1)

        if self.bias is not None:
            bias = self.bias[channels]
            output = output + bias

        output = self.activation(output)

        if self.residual:
            if output.shape[2] == x.shape[2]:
                output = output + x
            elif output.shape[2] == x.shape[2] * 2:
                output = output + torch.cat([x, x], dim=2)
            else:
                raise NotImplementedError("Not implemented")

        return output, channels


class Sequential(nn.Sequential):
    def forward(self, xx: Tuple[torch.Tensor, torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor]:
        for module in self:
            xx = module(xx)
        return xx


class AtomFitNet(nn.Module):
    def __init__(self,
                 num_channels: int,
                 network_size: List[int],
                 activate: str,
                 bias: bool,
                 resnet_dt: bool):
        super(AtomFitNet, self).__init__()

        self.bias = bias
        self.resnet_dt = resnet_dt

        layers = []
        for i in range(len(network_size) - 1):
            layers.append(Dense(num_channels=num_channels, in_features=network_size[i], out_features=network_size[i+1],
                                bias=bias, activate=activate, residual=resnet_dt)) # iterating through the neurons

        self.fitting_net = Sequential(*layers)

    def forward(self, x: torch.Tensor, channels: torch.Tensor) -> torch.Tensor:

        xx: Tuple[torch.Tensor, torch.Tensor] = (x, channels[0])
        output, _ = self.fitting_net(xx)
        return output

